from math import log, pi

from bhr.u_tube import UTube
from bhr.utilities import coth


class SingleUBorehole(UTube):
    def __init__(
        self,
        borehole_diameter: float,
        pipe_outer_diameter: float,
        pipe_dimension_ratio: float,
        length: float,
        shank_space: float,
        pipe_conductivity: float,
        grout_conductivity: float,
        soil_conductivity: float,
        fluid_type: str,
        fluid_concentration: float = 0,
    ):
        super().__init__(
            pipe_outer_diameter,
            pipe_dimension_ratio,
            length,
            shank_space,
            pipe_conductivity,
            fluid_type,
            fluid_concentration,
        )

        """
        Implementation for computing borehole thermal resistance for grouted single u-tube borehole.

        Relies primarily on the following references:

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Javed, S. & Spitler, J.D. Calculation of Borehole Thermal Resistance. In 'Advances in
        Ground-Source Heat Pump Systems,' pp. 84. Rees, S.J. ed. Cambridge, MA. Elsevier Ltd. 2016.

        :param borehole_diameter: borehole diameter, in m.
        :param pipe_outer_diameter: outer diameter of the pipe, in m.
        :param pipe_dimension_ratio: non-dimensional ratio of pipe diameter to pipe thickness.
        :param length: length of borehole from top to bottom, in m.
        :param shank_space: radial distance from the borehole center to the pipe center, in m.
        :param pipe_conductivity: pipe thermal conductivity, in W/m-K.
        :param grout_conductivity: grout thermal conductivity, in W/m-K.
        :param soil_conductivity: soil thermal conductivity, in W/m-K.
        :param fluid_type: fluid type. "ETHYLALCOHOL", "ETHYLENEGLYCOL", "METHYLALCOHOL",  "PROPYLENEGLYCOL", or "WATER"
        :param fluid_concentration: fractional concentration of antifreeze mixture, from 0-0.6.
        """

        # static parameters
        self.borehole_diameter = borehole_diameter
        self.grout_conductivity = grout_conductivity
        self.soil_conductivity = soil_conductivity
        self.theta_1 = 2 * self.shank_space / self.borehole_diameter
        self.theta_2 = self.borehole_diameter / self.pipe_outer_diameter
        self.theta_3 = 1 / (2 * self.theta_1 * self.theta_2)
        self.sigma = (self.grout_conductivity - self.soil_conductivity) / (
            self.grout_conductivity + self.soil_conductivity
        )
        self.bh_length = length
        self.two_pi_kg = 2 * pi * self.grout_conductivity

        # non-static parameters
        self.pipe_resist = None

    def update_beta(self, m_dot: float, temp: float) -> float:
        """
        Updates Beta coefficient.

        Javed, S. & Spitler, J.D. Calculation of Borehole Thermal Resistance. In 'Advances in
        Ground-Source Heat Pump Systems,' pp. 84. Rees, S.J. ed. Cambridge, MA. Elsevier Ltd. 2016.

        Eq: 3-47

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Eq: 14

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius
        """

        pipe_resist = self.calc_fluid_pipe_resist(m_dot, temp)
        self.pipe_resist = pipe_resist
        beta = self.two_pi_kg * pipe_resist

        return beta

    def calc_direct_coupling_resistance(self, m_dot: float, temp: float) -> tuple:
        r_a = self.calc_total_internal_bh_resistance(m_dot, temp)
        r_b = self.calc_local_bh_resistance(m_dot, temp)

        r_12 = (4 * r_a * r_b) / (4 * r_b - r_a)

        # reset if negative
        if r_12 < 0:
            r_12 = 70

        resist_bh_direct_coupling = r_12
        return resist_bh_direct_coupling, r_b

    def calc_local_bh_resistance(self, m_dot: float, temp: float) -> float:
        """
        Calculates the average thermal resistance of the borehole using the first-order multipole method.

        Resistance between the fluid in the U-tube(s) to the borehole wall (m-K/W)

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Equation 13

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: average thermal resistance, K/(W/m)
        """
        beta = self.update_beta(m_dot, temp)

        final_term_1 = log(self.theta_2 / (2 * self.theta_1 * (1 - self.theta_1**4) ** self.sigma))

        term_2_num = self.theta_3**2 * (1 - (4 * self.sigma * self.theta_1**4) / (1 - self.theta_1**4)) ** 2
        term_2_den_pt_1 = (1 + beta) / (1 - beta)
        term_2_den_pt_2 = self.theta_3**2 * (1 + (16 * self.sigma * self.theta_1**4) / (1 - self.theta_1**4) ** 2)
        term_2_den = term_2_den_pt_1 + term_2_den_pt_2
        final_term_2 = term_2_num / term_2_den

        resist_bh_ave = (1 / (4 * pi * self.grout_conductivity)) * (beta + final_term_1 - final_term_2)
        return resist_bh_ave

    def calc_total_internal_bh_resistance(self, m_dot: float, temp: float) -> float:
        """
        Calculates the total internal thermal resistance of the borehole using the first-order multipole method.

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Equation 26

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: total internal thermal resistance, K/(W/m)
        """
        beta = self.update_beta(m_dot, temp)

        term_1_num = (1 + self.theta_1**2) ** self.sigma
        term_1_den = self.theta_3 * (1 - self.theta_1**2) ** self.sigma
        final_term_1 = log(term_1_num / term_1_den)

        term_2_num = self.theta_3**2 * (1 - self.theta_1**4 + 4 * self.sigma * self.theta_1**2) ** 2
        term_2_den_pt_1 = (1 + beta) / (1 - beta) * (1 - self.theta_1**4) ** 2
        term_2_den_pt_2 = self.theta_3**2 * (1 - self.theta_1**4) ** 2
        term_2_den_pt_3 = 8 * self.sigma * self.theta_1**2 * self.theta_3**2 * (1 + self.theta_1**4)
        term_2_den = term_2_den_pt_1 - term_2_den_pt_2 + term_2_den_pt_3
        final_term_2 = term_2_num / term_2_den

        resist_bh_total_internal = 1 / (pi * self.grout_conductivity) * (beta + final_term_1 - final_term_2)

        return resist_bh_total_internal

    def calc_grout_resistance(self, m_dot: float, temp: float) -> float:
        """
        Calculates grout resistance. Use for validation.

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Eq: 3

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: grout resistance, K/(W-m)
        """

        if self.pipe_resist is None:
            raise ValueError("Pipe resistance has not been calculated yet.")

        resist_bh_grout = self.calc_local_bh_resistance(m_dot, temp) - self.pipe_resist / 2.0
        return resist_bh_grout

    def calc_effective_bh_resistance_uhf(self, m_dot: float, temp: float) -> float:
        """
        Calculates the effective thermal resistance of the borehole assuming a uniform heat flux.

        Javed, S. & Spitler, J.D. Calculation of Borehole Thermal Resistance. In 'Advances in
        Ground-Source Heat Pump Systems,' pp. 84. Rees, S.J. ed. Cambridge, MA. Elsevier Ltd. 2016.

        Eq: 3-67

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: effective thermal resistance, K/(W/m)
        """

        r_a = self.calc_total_internal_bh_resistance(m_dot, temp)
        r_b = self.calc_local_bh_resistance(m_dot, temp)

        pt_1 = 1 / (3 * r_a)
        pt_2 = (self.bh_length / (self.fluid.cp(temp) * m_dot)) ** 2
        resist_short_circuiting = pt_1 * pt_2

        resist_bh_effective_uhf = r_b + resist_short_circuiting
        return resist_bh_effective_uhf

    def calc_effective_bh_resistance_ubwt(self, m_dot: float, temp: float) -> float:
        """
        Calculates the effective thermal resistance of the borehole assuming a uniform borehole wall temperature.

        Javed, S. & Spitler, J.D. Calculation of Borehole Thermal Resistance. In 'Advances in
        Ground-Source Heat Pump Systems,' pp. 84. Rees, S.J. ed. Cambridge, MA. Elsevier Ltd. 2016.

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: effective thermal resistance, K/(W/m)
        """

        r_a = self.calc_total_internal_bh_resistance(m_dot, temp)  # R_a
        r_b = self.calc_local_bh_resistance(m_dot, temp)  # R_b
        r_v = self.bh_length / (m_dot * self.fluid.cp(temp))  # (K/(w/m)) thermal resistance factor
        n = r_v / (r_b * r_a) ** 0.5
        resist_bh_effective_ubt = r_b * n * coth(n)

        return resist_bh_effective_ubt



from enum import Enum, auto


class BoundaryCondition(Enum):
    UNIFORM_BOREHOLE_WALL_TEMP = auto()
    UNIFORM_HEAT_FLUX = auto()


class BoreholeType(Enum):
    SINGLE_U_TUBE = auto()
    DOUBLE_U_TUBE = auto()
    COAXIAL = auto()


class DoubleUPipeInletArrangement(Enum):
    ADJACENT = auto()
    DIAGONAL = auto()



from math import cosh, exp, sinh

from bhr.enums import BoundaryCondition


def set_boundary_condition_enum(bh_str: str) -> BoundaryCondition:
    boundary_condition_map = {bc.name: bc for bc in BoundaryCondition}
    if bh_str.upper() in boundary_condition_map:
        return boundary_condition_map[bh_str.upper()]

    raise ValueError(f"Invalid boundary condition: '{bh_str}'")


def inch_to_m(x_inch: float) -> float:
    """
    Convert inches to meters

    :param x_inch: value to convert, inches
    :return: float value, meters
    """
    return x_inch * 0.0254


def smoothing_function(x, x_low_limit, x_high_limit, y_low_limit, y_high_limit) -> float:
    """
    Sigmoid smoothing function

    https://en.wikipedia.org/wiki/Sigmoid_function

    :param x: independent variable
    :param x_low_limit: lower limit on x range
    :param x_high_limit: upper limit on x range
    :param y_low_limit: lower limit on y range
    :param y_high_limit: upper limit on y range
    :return: smoothed value between x_low_limit and x_high_limit. returns y_low_limit and y_high_limit
    below and above the x_low_limit and x_high_limit, respectively.
    """

    # check lower and upper bounds
    if x < x_low_limit:
        return y_low_limit

    if x > x_high_limit:
        return y_high_limit

    # interp x range to value between -5 to 5 for sigmoid function, which is picked to be
    # within approx. 1% of the desired end points of the interpolation range, 0 to 1.
    s_x_max = 5
    s_x_min = -5
    s_x = (x - x_low_limit) / (x_high_limit - x_low_limit) * (s_x_max - s_x_min) + s_x_min

    # get sigmoid function for interpolation between real y values
    s_y = 1 / (1 + exp(-s_x))

    # finally interpolate to the final value
    return s_y * (y_high_limit - y_low_limit) + y_low_limit


def coth(x):
    return cosh(x) / sinh(x)



from math import log as ln
from math import pi, sqrt

from bhr.enums import DoubleUPipeInletArrangement
from bhr.u_tube import UTube
from bhr.utilities import coth


class DoubleUTube(UTube):
    def __init__(
        self,
        borehole_diameter: float,
        pipe_outer_diameter: float,
        pipe_dimension_ratio: float,
        length: float,
        shank_space: float,
        pipe_conductivity: float,
        pipe_inlet_arrangement: str,
        grout_conductivity: float,
        soil_conductivity: float,
        fluid_type: str,
        fluid_concentration: float = 0,
    ):
        super().__init__(
            pipe_outer_diameter,
            pipe_dimension_ratio,
            length,
            shank_space,
            pipe_conductivity,
            fluid_type,
            fluid_concentration,
        )

        """
        Implementation for computing borehole thermal resistance for grouted boreholes with parallel double u-tubes.

        Relies primarily on the following references:

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Claesson, Johan, and Saqib Javed. 2019. “Explicit Multipole Formulas and Thermal Network Models
        for Calculating Thermal Resistances of Double U-Pipe Borehole Heat Exchangers.” Science and Technology for
        the Built Environment 25 (8): 980-92. doi:10.1080/23744731.2019.1620565.

        :param borehole_diameter: borehole diameter, in m.
        :param pipe_outer_diameter: outer diameter of the pipe, in m.
        :param pipe_dimension_ratio: non-dimensional ratio of pipe diameter to pipe thickness.
        :param length: length of borehole from top to bottom, in m.
        :param shank_space: radial distance from the borehole center to the pipe center, in m.
        :param pipe_conductivity: pipe thermal conductivity, in W/m-K.
        :param pipe_inlet_arrangement: arrangement of the pipe inlets. "ADJACENT", or "DIAGONAL"
        :param grout_conductivity: grout thermal conductivity, in W/m-K.
        :param soil_conductivity: soil thermal conductivity, in W/m-K.
        :param fluid_type: fluid type. "ETHYLALCOHOL", "ETHYLENEGLYCOL", "METHYLALCOHOL",  "PROPYLENEGLYCOL", or "WATER"
        :param fluid_concentration: fractional concentration of antifreeze mixture, from 0-0.6.
        """

        # static parameters
        self.borehole_diameter = borehole_diameter
        self.grout_conductivity = grout_conductivity
        self.borehole_radius = borehole_diameter / 2  # radius of borehole (m)
        self.pipe_radius = pipe_outer_diameter / 2  # pipe outer radius (m)

        if pipe_inlet_arrangement == DoubleUPipeInletArrangement.ADJACENT.name:
            self.pipe_inlet_arrangement = DoubleUPipeInletArrangement.ADJACENT
        elif pipe_inlet_arrangement == DoubleUPipeInletArrangement.DIAGONAL.name:
            self.pipe_inlet_arrangement = DoubleUPipeInletArrangement.DIAGONAL
        else:
            msg = (
                f"Invalid pipe_inlet_arrangement. Use one of the allowed values: "
                f"{', '.join(map(str, DoubleUPipeInletArrangement._member_names_))}"
            )
            raise AssertionError(msg)

        self.bh_length = length  # length of borehole (m)
        self.grout_conductivity = grout_conductivity  # W/(m-K)
        self.soil_conductivity = soil_conductivity  # W/(m-K)

        # (m) radial distance between centers of symmetrically placed pipes and borehole center (rc)
        self.pipe_centers_radius = shank_space

        # thermal conductivity ratio, dimensionless
        self.sigma = (self.grout_conductivity - self.soil_conductivity) / (
            self.grout_conductivity + self.soil_conductivity
        )

        # Check if shank spacing realistic
        lower_shank_space_limit = sqrt(pipe_outer_diameter**2 / 2)
        upper_shank_space_limit = 0.5 * (borehole_diameter - pipe_outer_diameter)
        if shank_space < lower_shank_space_limit:
            msg = (
                "Shank spacing is too small and must be greater than the 2 pipe radii to prevent "
                "pipes from overlapping."
            )
            raise AssertionError(msg)
        elif shank_space > upper_shank_space_limit:
            msg = (
                "Shank spacing is too large and must be less than the borehole radius minus the pipe radius "
                "to prevent pipes from extending beyond the borehole wall."
            )
            raise AssertionError(msg)

        # static parameters - calc_bh_resist_local
        self.p_pc = self.pipe_radius**2 / (4 * self.pipe_centers_radius**2)
        self.p_c = self.pipe_centers_radius**2 / (self.borehole_radius**8 - self.pipe_centers_radius**8) ** 0.25
        self.p_b = self.borehole_radius**2 / (self.borehole_radius**8 - self.pipe_centers_radius**8) ** 0.25
        self.eight_pi_kg = 8 * pi * self.grout_conductivity
        self.b_2 = ln(self.borehole_radius**4 / (4 * self.pipe_radius * self.pipe_centers_radius**3))
        self.b_3 = ln(self.borehole_radius**8 / (self.borehole_radius**8 - self.pipe_centers_radius**8))

        # static parameter - calc_internal_resist
        self.two_pi_kg = 2 * pi * self.grout_conductivity
        self.c_1 = self.pipe_centers_radius / self.pipe_radius
        c_2 = self.borehole_radius**4 + self.pipe_centers_radius**4
        c_3 = self.borehole_radius**4 - self.pipe_centers_radius**4
        self.ln_c2_c3 = ln(c_2 / c_3)
        self.c_4 = self.p_c**2 * self.p_b**2
        self.c_5 = self.p_c**2 * self.p_b**6 + self.p_c**6 * self.p_b**2
        d_2 = self.borehole_radius**2 + self.pipe_centers_radius**2
        d_3 = self.borehole_radius**2 - self.pipe_centers_radius**2
        self.ln_d2_d3 = ln(d_2 / d_3)
        self.d_4 = 3 * self.p_c**3 * self.p_b**5 + self.p_c**7 * self.p_b
        self.d_5 = self.p_c * self.p_b**7 + 3 * self.p_c**5 * self.p_b**3

        # non-static parameters
        self.pipe_resist: float | None = None

    def update_b1(self, m_dot_per_u_tube: float, temperature: float) -> float:
        """
        Updates b1 coefficient.

        Javed, S. & Spitler, J.D. Calculation of Borehole Thermal Resistance. In 'Advances in
        Ground-Source Heat Pump Systems,' pp. 84. Rees, S.J. ed. Cambridge, MA. Elsevier Ltd. 2016.

        Eq: 3-47

        Javed, S. & Spitler, J.D. 2017. 'Accuracy of Borehole Thermal Resistance Calculation Methods
        for Grouted Single U-tube Ground Heat Exchangers.' Applied Energy.187:790-806.

        Eq: 14

        :param m_dot_per_u_tube: mass flow rate in each u-tube, kg/s
        :param temperature: temperature, Celsius

        :return: b1: a ratio of (1-beta)/(1+beta), dependent on pipe resistance
                    & grout conductivity, dimensionless
        """

        pipe_resist = self.calc_fluid_pipe_resist(m_dot_per_u_tube, temperature)
        self.pipe_resist = pipe_resist
        beta = 2 * pi * self.grout_conductivity * pipe_resist
        b1 = (1 - beta) / (1 + beta)  # dimensionless parameter

        return b1

    def calc_bh_resist_local(self, m_dot_per_u_tube: float, temperature: float) -> float:
        """
        Calculates tube-to-borehole resistance (aka local borehole resistance).

        Claesson, Johan, and Saqib Javed. 2019. “Explicit Multipole Formulas
        and Thermal Network Models for Calculating Thermal Resistances of
        Double U-Pipe Borehole Heat Exchangers.” Science and Technology for
        the Built Environment 25 (8): 980-92. doi:10.1080/23744731.2019.1620565.

        Eq: 13 & 14

        :param m_dot_per_u_tube: mass flow rate in each u-tube, kg/s
        :param temperature: temperature, Celsius

        :return: borehole_resist_local: local borehole resistance, K/(W/m)
        """

        b1 = self.update_b1(m_dot_per_u_tube, temperature)

        if self.pipe_resist is None:
            raise ValueError("Pipe resistance has not been calculated yet.")

        # --Borehole resistance, 0th order [K/(W/m)]--
        rb0 = self.pipe_resist / 4 + 1 / self.eight_pi_kg * (self.b_2 + self.sigma * self.b_3)

        # --Borehole resistance, 1st order [K/(W/m)]--
        borehole_resist_local = rb0 - 1 / self.eight_pi_kg * (
            b1 * self.p_pc * (3 - 8 * self.sigma * self.p_c**4) ** 2
        ) / (1 + b1 * self.p_pc * (5 + 64 * self.sigma * self.p_c**4 * self.p_b**4))

        return borehole_resist_local

    def calc_internal_resist(self, m_dot_per_u_tube: float, temperature: float) -> float:
        """
        Calculates tube-to-tube resistance (aka internal resistance).

        Claesson, Johan, and Saqib Javed. 2019. “Explicit Multipole Formulas
        and Thermal Network Models for Calculating Thermal Resistances of
        Double U-Pipe Borehole Heat Exchangers.” Science and Technology for
        the Built Environment 25 (8): 980-92. doi:10.1080/23744731.2019.1620565.

        Eq: 18, 19, 22, 23

        :param m_dot_per_u_tube: mass flow rate in each u-tube, kg/s
        :param temperature: temperature, Celsius

        :return: internal_resist: local internal resistance, K/(W/m)
        """

        b1 = self.update_b1(m_dot_per_u_tube, temperature)

        if self.pipe_resist is None:
            raise ValueError("Pipe resistance has not been calculated yet.")

        if self.pipe_inlet_arrangement == DoubleUPipeInletArrangement.DIAGONAL:
            # 0th order
            ra0 = 2 * self.pipe_resist + 2 / self.two_pi_kg * (ln(self.c_1) + self.sigma * self.ln_c2_c3)

            # 1st order
            internal_resist = ra0 - 2 / self.two_pi_kg * (b1 * self.p_pc * (1 + 8 * self.sigma * self.c_4) ** 2) / (
                1 - b1 * self.p_pc * (3 - 32 * self.sigma * self.c_5)
            )

            return internal_resist

        elif self.pipe_inlet_arrangement == DoubleUPipeInletArrangement.ADJACENT:
            # 0th order
            ra0 = 2 * self.pipe_resist + 2 / self.two_pi_kg * (ln(2 * self.c_1) + self.sigma * self.ln_d2_d3)

            # 1st order
            matrix_element_11 = 1 + 16 * b1 * self.sigma * self.p_pc * self.d_4
            matrix_element_22 = -1 - 16 * b1 * self.sigma * self.p_pc * self.d_5
            matrix_element_21 = b1 * self.p_pc
            vector_1 = 1 - 8 * self.sigma * self.p_c**3 * self.p_b
            vector_2 = 3 + 8 * self.sigma * self.p_c * self.p_b**3

            internal_resist = ra0 + 2 / self.two_pi_kg * b1 * self.p_pc / 2 * (
                vector_2**2 * matrix_element_11
                - 2 * vector_1 * vector_2 * matrix_element_21
                - vector_1**2 * matrix_element_22
            ) / (matrix_element_11 * matrix_element_22 + matrix_element_21**2)

            return internal_resist

        raise AssertionError("Developer error. Invalid pipe inlet arrangement.")

    def calc_effective_bh_resistance_uhf(self, m_dot: float, temp: float) -> float:
        """
        Calculates effective borehole resistance for uniform heat flux along the borehole.

        Claesson, Johan, and Saqib Javed. 2019. “Explicit Multipole Formulas
        and Thermal Network Models for Calculating Thermal Resistances of
        Double U-Pipe Borehole Heat Exchangers.” Science and Technology for
        the Built Environment 25 (8): 980-92. doi:10.1080/23744731.2019.1620565.

        Eq: 44

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: effective_bhr_uhf: effective borehole resistance under uniform heat flux boundary conditions [K/(W/m)]
        """

        m_dot_per_u_tube = m_dot / 2
        internal_resist = self.calc_internal_resist(m_dot_per_u_tube, temp)
        borehole_resist_local = self.calc_bh_resist_local(m_dot_per_u_tube, temp)
        rv = self.bh_length / (self.fluid.cp(temp) * m_dot_per_u_tube)  # (K/(w/m)) thermal resistance factor

        effective_bhr_uhf = borehole_resist_local + rv**2 / (6 * internal_resist)

        return effective_bhr_uhf

    def calc_effective_bh_resistance_ubwt(self, m_dot: float, temp: float) -> float:
        """
        Calculates effective borehole resistance for uniform borehole wall temperature.

        Claesson, Johan, and Saqib Javed. 2019. “Explicit Multipole Formulas
        and Thermal Network Models for Calculating Thermal Resistances of
        Double U-Pipe Borehole Heat Exchangers.” Science and Technology for
        the Built Environment 25 (8): 980-92. doi:10.1080/23744731.2019.1620565.

        Eq: 46

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, Celsius

        :return: effective_bhr_ubwt: effective borehole resistance for uniform borehole wall temperature
                                    boundary condition [K/(W/m)]
        """

        m_dot_per_u_tube = m_dot / 2
        internal_resist = self.calc_internal_resist(m_dot_per_u_tube, temp)
        borehole_resist_local = self.calc_bh_resist_local(m_dot_per_u_tube, temp)

        rv = self.bh_length / (self.fluid.cp(temp) * m_dot_per_u_tube)  # (K/(w/m)) thermal resistance factor
        n = rv / (2 * borehole_resist_local * internal_resist) ** 0.5
        effective_bhr_ubwt = borehole_resist_local * n * coth(n)

        return effective_bhr_ubwt



from math import log, pi

from bhr.fluid import get_fluid
from bhr.utilities import inch_to_m, smoothing_function


class Pipe:
    def __init__(
        self,
        pipe_outer_diameter: float,
        pipe_dimension_ratio: float,
        pipe_length: float,
        pipe_conductivity: float,
        fluid_type: str,
        fluid_concentration: float = 0,
    ):
        self.fluid = get_fluid(fluid_type, fluid_concentration)

        # ratio of outer diameter to wall thickness
        self.dimension_ratio = pipe_dimension_ratio

        # set diameters and thickness
        self.pipe_outer_diameter = pipe_outer_diameter
        self.pipe_inner_diameter = self.pipe_outer_diameter * (1 - 2 / self.dimension_ratio)
        self.thickness = self.pipe_outer_diameter / self.dimension_ratio

        # set length
        self.pipe_length = pipe_length

        # set physical properties
        self.pipe_conductivity = pipe_conductivity

        # compute cross-sectional areas
        self.area_cr_inner = pi / 4 * self.pipe_inner_diameter**2
        self.area_cr_outer = pi / 4 * self.pipe_outer_diameter**2
        self.area_cr_pipe = self.area_cr_outer - self.area_cr_inner

        # compute surface areas
        self.area_s_inner = pi * self.pipe_inner_diameter * self.pipe_length
        self.area_s_outer = pi * self.pipe_outer_diameter * self.pipe_length

        # compute volumes
        self.total_vol = self.area_cr_outer * self.pipe_length
        self.fluid_vol = self.area_cr_inner * self.pipe_length
        self.pipe_wall_vol = self.area_cr_pipe * self.pipe_length

    @staticmethod
    def get_inner_dia(outer_dia: float, dimension_ratio: float) -> float:
        return outer_dia * (1 - 2 / dimension_ratio)

    def get_pipe_diameters_imperial(self, nominal_pipe_size_inches: float, dimension_ratio: float):
        if nominal_pipe_size_inches == 0.75:
            outer_dia = 1.05
        elif nominal_pipe_size_inches == 1.0:
            outer_dia = 1.315
        elif nominal_pipe_size_inches == 1.25:
            outer_dia = 1.66
        elif nominal_pipe_size_inches == 1.5:
            outer_dia = 1.9
        elif nominal_pipe_size_inches == 2.0:
            outer_dia = 2.375
        elif nominal_pipe_size_inches == 3.0:
            outer_dia = 3.5
        elif nominal_pipe_size_inches == 4.0:
            outer_dia = 4.5
        elif nominal_pipe_size_inches == 6.0:
            outer_dia = 6.625
        elif nominal_pipe_size_inches == 8.0:
            outer_dia = 8.625
        else:
            raise ValueError("Unsupported pipe size")

        return inch_to_m(self.get_inner_dia(outer_dia, dimension_ratio)), inch_to_m(outer_dia)

    def mdot_to_vdot(self, m_dot: float, temp: float) -> float:
        """
        Computes volumetric flow rate based on mass flow rate.

        :param m_dot: mass flow rate, in kg/s
        :param temp: temperature, in C
        :return: volumetric flow rate, in m3/s
        """

        return m_dot / self.fluid.density(temp)

    def mdot_to_re(self, m_dot: float, temp: float) -> float:
        """
        Computes Reynolds number based on mass flow rate.

        :param m_dot: mass flow rate, in kg/s
        :param temp: temperature, in C
        :return: Reynolds number, dimensionless
        """

        return 4 * m_dot / (self.fluid.mu(temp) * pi * self.pipe_inner_diameter)

    def mdot_to_velocity(self, m_dot: float, temp: float) -> float:
        """
        Computes velocity based on mass flow rate.

        :param m_dot: mass flow rate in, kg/s
        :param temp: temperature, in C
        :return: velocity, in m/s
        """

        return m_dot / (self.area_cr_inner * self.fluid.density(temp))

    def friction_factor(self, re: float) -> float:
        """
        Calculates the friction factor in smooth tubes

        Petukhov, B.S. 1970. 'Heat transfer and friction in turbulent pipe flow with variable physical properties.'
        In Advances in Heat Transfer, ed. T.F. Irvine and J.P. Hartnett, Vol. 6. New York Academic Press.

        :param re: Reynolds number, dimensionless
        """

        # limits picked be within about 1% of actual values
        low_reynolds = 2000
        high_reynolds = 4000

        if re < low_reynolds:
            return self.laminar_friction_factor(re)
        if re > high_reynolds:
            return self.turbulent_friction_factor(re)

        # pure laminar flow
        f_low = self.laminar_friction_factor(re)

        # pure turbulent flow
        f_high = self.turbulent_friction_factor(re)

        return smoothing_function(re, low_reynolds, high_reynolds, f_low, f_high)

    @staticmethod
    def laminar_friction_factor(re: float):
        """
        Laminar friction factor

        :param re: Reynolds number
        :return: friction factor
        """

        return 64.0 / re

    @staticmethod
    def turbulent_friction_factor(re: float):
        """
        Turbulent friction factor

        Petukhov, B. S. (1970). Advances in Heat Transfer, volume 6, Heat transfer and
        friction in turbulent pipe flow with variable physical properties, pages 503-564.
        Academic Press, Inc., New York, NY.

        :param re: Reynolds number
        :return: friction factor
        """

        return (0.79 * log(re) - 1.64) ** (-2.0)

    def pressure_loss(self, m_dot: float, temp: float) -> float:
        """
        Pressure loss in straight pipe

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, C
        :return: pressure loss, Pa
        """

        if m_dot <= 0:
            return 0

        re = self.mdot_to_re(m_dot, temp)
        term_1 = self.friction_factor(re) * self.pipe_length / self.pipe_inner_diameter
        term_2 = (self.fluid.density(temp) * self.mdot_to_velocity(m_dot, temp) ** 2) / 2

        return term_1 * term_2

    def pressure_loss_v_dot(self, v_dot: float, temp: float) -> float:
        """
        Pressure loss in a straight pipe

        :param v_dot: volume flow rate, m3/s
        :param temp: temperature, C
        :return: pressure loss, Pa
        """

        m_dot = self.fluid.density(temp) * v_dot
        return self.pressure_loss(m_dot, temp)

    @staticmethod
    def laminar_nusselt():
        """
        Laminar Nusselt number for smooth pipes

        mean(4.36, 3.66)
        :return: Nusselt number
        """
        return 4.01

    def turbulent_nusselt(self, re: float, temp: float):
        """
        Turbulent Nusselt number for smooth pipes

        Gnielinski, V. 1976. 'New equations for heat and mass transfer in turbulent pipe and channel flow.'
        International Chemical Engineering 16(1976), pp. 359-368.

        :param re: Reynolds number
        :param temp: temperature, C
        :return: Nusselt number
        """

        f = self.friction_factor(re)
        pr = self.fluid.prandtl(temp)
        return (f / 8) * (re - 1000) * pr / (1 + 12.7 * (f / 8) ** 0.5 * (pr ** (2 / 3) - 1))

    def calc_cond_resist(self) -> float:
        """
        Calculates the pipe radial conduction thermal resistance, in [K/(W/m)].

        Javed, S. and Spitler, J.D. 2017. 'Accuracy of borehole thermal resistance calculation methods
        for grouted single U-tube ground heat exchangers.' Applied Energy. 187: 790-806.

        :return: conduction resistance, K/(W/m)
        """

        return log(self.pipe_outer_diameter / self.pipe_inner_diameter) / (2 * pi * self.pipe_conductivity)

    def calc_conv_resist(self, m_dot: float, temp: float) -> float:
        """
        Calculates the pipe internal convection thermal resistance, in [k/(W/m)]

        Gnielinski, V. 1976. 'New equations for heat and mass transfer in turbulent pipe and channel flow.'
        International Chemical Engineering 16(1976), pp. 359-368.

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, C
        :return: convection resistance, K/(W/m)
        """

        low_reynolds = 2000
        high_reynolds = 4000

        re = self.mdot_to_re(m_dot, temp)

        if re < low_reynolds:
            nu = self.laminar_nusselt()
        elif low_reynolds <= re < high_reynolds:
            nu_low = self.laminar_nusselt()
            nu_high = self.turbulent_nusselt(high_reynolds, temp)
            nu = smoothing_function(re, low_reynolds, high_reynolds, nu_low, nu_high)
        else:
            nu = self.turbulent_nusselt(re, temp)

        return 1 / (nu * pi * self.fluid.k(temp))

    def calc_fluid_pipe_resist(self, m_dot: float, temp: float):
        """
        Calculates the combined convection and conduction pipe resistance

        Javed, S. and Spitler, J.D. 2017. 'Accuracy of borehole thermal resistance calculation methods
        for grouted single U-tube ground heat exchangers.' Applied Energy. 187: 790-806.

        Equation 3

        :param m_dot: mass flow rate, kg/s
        :param temp: temperature, C
        :return: combined convection and conduction pipe resistance, K/(W/m)
        """

        return self.calc_conv_resist(m_dot, temp) + self.calc_cond_resist()
